'''
Created on 22 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess

from t4gpd.commons.crossroads_identification.CrossroadRecognitionLib import CrossroadRecognitionLib


class CrossroadRecognition(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, method, patternsGdf, patternIdFieldName='gid', nRays=64, maxRayLength=None):
        '''
        Constructor
        '''
        if not isinstance(patternsGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(patternsGdf, 'GeoDataFrame')

        self.recognEngine = CrossroadRecognitionLib(method=method, nRays=nRays, maxRayLength=maxRayLength)
        self.patternSignatures = self.recognEngine.signatures(patternsGdf, patternIdFieldName, isAPatternLayer=True)

    def runWithArgs(self, row):
        geom = row.geometry

        raylens = GeomLib.fromMultiLineStringToLengths(geom)
        shapeSignature = self.recognEngine.signature(raylens)
        return self.recognEngine.nearestPattern(shapeSignature, self.patternSignatures)



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from numpy import sqrt
from t4gpd.commons.GeoProcess import GeoProcess

from t4gpd.commons.crossroads_generation.SequenceRadii import SequenceRadii
from t4gpd.commons.crossroads_generation.SequencesGeneration import SequencesGeneration


class STCrossroadsGeneration(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, nbranchs, length, width, mirror=False, withBranchs=True, withSectors=True,
                 crs='EPSG:2154', magnitude=2.5):
        '''
        Constructor
        '''
        self.nbranchs = nbranchs
        self.sequenceRadii = SequenceRadii(nbranchs, width, length)
        self.dictOfSequences = SequencesGeneration(nbranchs, mirror, withBranchs, withSectors).run()
        self.width = width
        self.offset = magnitude * length

        nseq = len(self.dictOfSequences)
        sqrt_nseq = int(sqrt(nseq))
        if (abs(nseq - sqrt_nseq * sqrt_nseq) < abs(nseq - (sqrt_nseq + 1) * (sqrt_nseq + 1))):
            self.ncols = sqrt_nseq
        else:
            self.ncols = sqrt_nseq + 1

        self.crs = crs

    def run(self):
        rows = list()
        for i, (gid, seq) in enumerate(self.dictOfSequences.items()):
            xoffset = self.offset * (i % self.ncols)
            yoffset = -self.offset * int(i / self.ncols)
            geom = seq.asPolygon(self.sequenceRadii, self.width, [xoffset, yoffset])
            row = dict({'gid': gid,
                        'vpoint_x': xoffset,
                        'vpoint_y': yoffset,
                        'model': seq.getMinModel(),
                        'sequence': str(seq),
                        'geometry': geom})
            rows.append(row)

        return GeoDataFrame(rows, crs=self.crs)



'''
Created on 15 sept. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from _warnings import warn

from pandas import read_csv
from t4gpd.commons.GeoProcess import GeoProcess


class KestrelReader(GeoProcess):
    '''
    classdocs
    '''
    FIELD_NAMES = [
        'timestamp', 'Tair', 'WetBulbTemperature', 'GlobeTemperature', 'RH',
        'BarometricPressure', 'Altitude', 'StationPressure', 'WindSpeed', 'HeatStressIdx',
        'DewPoint', 'DensityAltitude', 'CrosswindSpeed', 'HeadwindSpeed', 'CompassMD',
        'NAWetBulbTemperature', 'WindDirection', 'TWL', 'WBGT', 'WindChill', 'DataType'] 
    
    def __init__(self, inputFile):
        '''
        Constructor
        '''
        self.inputFile = inputFile

    def run(self):
        # https://kestrelinstruments.com/kestrel-5400-heat-stress-tracker
        df = read_csv(self.inputFile, decimal='.', header=None,
                      names=self.FIELD_NAMES, na_values='--', sep=',',
                      skiprows=5, parse_dates=[0], usecols=range(21))

        dfUnits = read_csv(self.inputFile, header=None, names=self.FIELD_NAMES,
                           nrows=1, sep=',', skiprows=4, usecols=range(21))

        if (('km/h' == dfUnits.WindSpeed.squeeze().lower()) or 
            ('km/h' == dfUnits.CrosswindSpeed.squeeze().lower()) or 
            ('km/h' == dfUnits.HeadwindSpeed.squeeze()).lower()):
            warn('Speed in km/h')
            # CONVERT km/h INTO m/s
            for fieldname in ['WindSpeed', 'CrosswindSpeed', 'HeadwindSpeed']:
                if ('km/h' == dfUnits[fieldname].squeeze().lower()):
                    df[fieldname] /= 3.6

        df['sensorFamily'] = 'Kestrel-5400'

        return df

'''
# inputFile = '/home/tleduc/prj/nm-ilots-frais/terrain/220713/HEAT_-_2414248_7_13_22_18_36_20.csv'
inputFile = '/home/tleduc/prj/nm-ilots-frais/terrain/220711/HEAT_-_2414248_7_11_22_18_55_00.csv'
df = KestrelReader(inputFile).run()
print(df.head(5))
'''



'''
Created on 12 oct. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import timedelta
import string

from numpy.random import choice
from pandas import DataFrame, merge_asof
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class JoinByTimeDistance(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, dfImu, dfMob, left_on='timestamp', right_on='timestamp',
                 tolerance=timedelta(minutes=1)):
        '''
        Constructor
        '''
        if not isinstance(dfImu, DataFrame):
            raise IllegalArgumentTypeException(dfImu, 'DataFrame')
        self.dfImu = dfImu

        if not isinstance(dfMob, DataFrame):
            raise IllegalArgumentTypeException(dfMob, 'DataFrame')
        self.dfMob = dfMob

        if not left_on in dfImu:
            raise IllegalArgumentTypeException(left_on, 'is not a valid field name')
        self.left_on = left_on

        if not right_on in dfMob:
            raise IllegalArgumentTypeException(right_on, 'is not a valid field name')
        self.right_on = right_on

        if not isinstance(tolerance, timedelta):
            raise IllegalArgumentTypeException(tolerance, 'timedelta')
        self.tolerance = tolerance

        assert dfImu[left_on].is_monotonic_increasing, f'dfImu.{left_on} is not monotonic increasing!'
        assert dfMob[right_on].is_monotonic_increasing, f'dfMob.{right_on} is not monotonic increasing!'

    def __getRndFieldName(self):
        return ''.join(choice(list(string.ascii_letters), size=10))

    def run(self):
        tmpFieldName = self.__getRndFieldName()
        self.dfMob.rename(columns={self.right_on: tmpFieldName}, inplace=True)

        # ~ https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.merge_asof.html
        gdf = merge_asof(self.dfImu, self.dfMob, left_on=self.left_on,
                         right_on=tmpFieldName, direction='nearest',
                         tolerance=self.tolerance)

        self.dfMob.rename(columns={tmpFieldName: self.right_on}, inplace=True)

        gdf['timedelta_in_s'] = list(zip(gdf[self.left_on], gdf[tmpFieldName]))
        gdf.timedelta_in_s = gdf.timedelta_in_s.apply(lambda t: (t[0] - t[1]).total_seconds()) 

        gdf.drop(columns=[tmpFieldName], inplace=True)
        return gdf

'''
from t4gpd.picoclim.MetrologicalCampaignReader import MetrologicalCampaignReader
from t4gpd.picoclim.SnapImuOnTrackUsingWaypoints import SnapImuOnTrackUsingWaypoints

# dirName = '/home/tleduc/prj/nm-ilots-frais/terrain/220711'
dirName = '/home/tleduc/prj/nm-ilots-frais/terrain/220713'
tracks, waypoints, dfImu, dfMob, dfStat1, dfStat2 = MetrologicalCampaignReader(dirName).run()

imu = SnapImuOnTrackUsingWaypoints(dfImu, tracks, waypoints).run()

dfImuMob = JoinByTimeDistance(imu, dfMob, left_on='timestamp', right_on='timestamp').run()
'''



'''
Created on 11 oct. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import isnan, nan
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class SnapImuOnTrackUsingWaypoints(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, dfImu, tracks, waypoints):
        '''
        Constructor
        '''
        if not isinstance(dfImu, GeoDataFrame):
            raise IllegalArgumentTypeException(dfImu, 'GeoDataFrame')
        self.dfImu = dfImu

        if not isinstance(tracks, GeoDataFrame):
            raise IllegalArgumentTypeException(tracks, 'GeoDataFrame')
        self.tracks = tracks

        if not isinstance(waypoints, GeoDataFrame):
            raise IllegalArgumentTypeException(waypoints, 'GeoDataFrame')
        self.waypoints = waypoints

        assert GeoDataFrameLib.shareTheSameCrs(tracks, waypoints), 'tracks and waypoints must share the same crs!'

    def run(self):
        maxTagNames = self.waypoints.maxTagName.unique()
        imu = self.dfImu.copy(deep=True)
        imu.to_crs(self.tracks.crs, inplace=True)

        # ==============================================================
        # 1st STEP: IDENTIFY THE TRACK TO PROJECT ON
        imu['track'], imu['curv_absc'], imu['n_curv_absc'] = None, None, None

        prev, cnt1, rows, cnts = nan, nan, [], []
        for cnt2, (_, row) in enumerate(imu.iterrows()):
            curr = row.TagName
            if isnan(prev):
                prev, cnt1 = (nan, nan) if isnan(curr) else (curr, 0)
                row.track, row.curv_absc = prev, cnt1
                if not isnan(curr):
                    cnts.append(cnt2)

            else:
                if isnan(curr):
                    cnts.append(cnt2)
                    cnt1 += 1
                    row.track, row.curv_absc = prev, cnt1
                else:

                    row.track, row.curv_absc = curr, 0
                    row.n_curv_absc = cnt1 + 1
                    for _cnt in cnts:
                        rows[_cnt].n_curv_absc = cnt1 + 1
                    cnts = []

                    if curr in maxTagNames:
                        prev, cnt1 = nan, nan
                    else:
                        prev, cnt1 = curr, 0

            rows.append(row)
        imu = GeoDataFrame(rows, crs=imu.crs)

        imu.curv_absc = list(zip(imu.track, imu.curv_absc, imu.n_curv_absc))
        imu.track = imu.track.apply(lambda v: v if isnan(v) else int(v // 100))

        # ==============================================================
        # 2nd STEP: ASSESS THE CURVILINEAR ABSCISSA OF THE PROJECTED POINT
        wpDict = self.waypoints.set_index('id').to_dict(orient='index')

        foo = lambda t: (None if isnan(t[0]) else (
            wpDict[t[0]]['curv_absc'] + (t[1] * wpDict[t[0]]['delta_curv_absc']) / t[2]))
        imu.curv_absc = imu.curv_absc.apply(lambda t: foo(t))

        for tn in maxTagNames:
            imu.loc[ imu[imu.TagName == tn].index, 'curv_absc' ] = 1.0

        # ==============================================================
        # 3rd STEP: ASSESS THE COORDINATES OF THE PROJECTED POINT
        trDict = self.tracks.set_index('id').to_dict(orient='index')

        imu['snap_geometry'] = list(zip(imu.track, imu.curv_absc))

        foo = lambda t: (None if isnan(t[0]) else (
            trDict[t[0]]['geometry'].interpolate(t[1], normalized=True)
            ))
        imu.snap_geometry = imu.snap_geometry.apply(lambda t: foo(t))

        # ==============================================================
        # 4th STEP: ASSESS THE DISTANCE (DRIFT) BETWEEN THE GNSS POINT AND 
        # ITS CORRESPONDING PROJECTED POINT
        imu.rename(columns={'geometry': 'gnss_geom', 'snap_geometry': 'geometry'}, inplace=True)
        imu.drop(columns=['n_curv_absc'], inplace=True)

        imu['drift'] = list(zip(imu.gnss_geom, imu.geometry))
        imu.drift = imu.drift.apply(lambda t:
                                    nan if ((t[0] is None) or (t[1] is None))
                                    else t[0].distance(t[1]))

        imu.gnss_geom = imu.gnss_geom.apply(lambda g: g.wkt)
        imu.set_crs(self.tracks.crs, inplace=True)

        # imu.drop(columns=['n_curv_absc'], inplace=True)

        imu.drop(index=imu[imu.geometry.isna()].index, inplace=True)
        imu.reset_index(drop=True, inplace=True)

        return imu

'''
from t4gpd.picoclim.MetrologicalCampaignReader import MetrologicalCampaignReader

# dirName = '/home/tleduc/prj/nm-ilots-frais/terrain/220711'
dirName = '/home/tleduc/prj/nm-ilots-frais/terrain/220713'
tracks, waypoints, dfImu, dfMob, dfStat1, dfStat2 = MetrologicalCampaignReader(dirName).run()

imu = SnapImuOnTrackUsingWaypoints(dfImu, tracks, waypoints).run()
# imu.drop(columns=['timestamp', 'X', 'Y', 'Z', 'Distance', 'degree', 'latitude',
#                   'longitude', 'GpsAccuracy', 'indoor_outdoor_flag'], inplace=True)
# imu.to_csv('/home/tleduc/prj/nm-ilots-frais/a.csv', index=False)
# imu.to_file('/tmp/imu.gpkg')
'''



'''
Created on 11 oct. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from os import access, R_OK
from os.path import isfile

from fiona import listlayers
from geopandas import read_file
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class GpkgLoader(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gpkgInputFile, layernames=None, bbox=None, mask=None):
        '''
        Constructor
        '''
        if not (isfile(gpkgInputFile) and access(gpkgInputFile, R_OK)):
            raise IllegalArgumentTypeException(gpkgInputFile, 'readable filename')
        self.gpkgInputFile = gpkgInputFile

        self.layernames = layernames
        self.bbox = bbox
        self.mask = mask

    def listlayers(self):
        return listlayers(self.gpkgInputFile)

    def run(self):
        layernames = listlayers(self.gpkgInputFile) if self.layernames is None else self.layernames

        result = {}
        for layername in layernames:
            df = read_file(self.gpkgInputFile, layer=layername, bbox=self.bbox, mask=self.mask)
            result[layername] = df
        return result



'''
Created on 19 sept. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import apply_along_axis, asarray, linspace
from shapely.coords import CoordinateSequence
from shapely.geometry import LinearRing, LineString, Point, Polygon
from sklearn.preprocessing import normalize
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.PointsDensifierLib3D import PointsDensifierLib3D


class SphericalProjectionLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def __projectionOnTheUnitSphere(vpCoord, coords):
        assert isinstance(vpCoord, CoordinateSequence), 'vpCoord is expected to be a CoordinateSequence!'
        assert 1 == len(vpCoord), 'vpCoord is expected to be a CoordinateSequence of a single Point!'
        assert 3 == len(vpCoord[0]), 'vpCoord is expected to be a 3D CoordinateSequence!'
        assert isinstance(coords, CoordinateSequence), 'coords is expected to be a CoordinateSequence!'

        vp = asarray(vpCoord)
        c = asarray(coords)
        nrows, ncols = c.shape
        assert 3 == ncols, 'coords is expected to be a 3D CoordinateSequence!'

        vp = vp.repeat(nrows, axis=0)
        prj = c - vp

        __filter = lambda row: (row[0], row[1], max(vpCoord[0][2], row[2]))
        return apply_along_axis(__filter, 1, vp + normalize(prj, axis=1, norm='l2'))

    @staticmethod
    def __stereographicProjectionOnTheUnitDisc(vpCoord, coords):
        prjCoords = SphericalProjectionLib.__projectionOnTheUnitSphere(vpCoord, coords)

        vp = asarray(vpCoord)
        c = asarray(prjCoords)
        nrows, _ = c.shape

        vp = vp.repeat(nrows, axis=0)
        prj = c - vp
        __stereo = lambda row: (row[0] / (1 + row[2]), row[1] / (1 + row[2]), 0)
        return vp + apply_along_axis(__stereo, 1, prj)

    @staticmethod
    def __orthogonalProjectionOnTheUnitDisc(vpCoord, coords):
        prjCoords = SphericalProjectionLib.__projectionOnTheUnitSphere(vpCoord, coords)

        vp = asarray(vpCoord)
        c = asarray(prjCoords)
        nrows, _ = c.shape

        vp = vp.repeat(nrows, axis=0)
        prj = c - vp
        __ortho = lambda row: (row[0], row[1], 0)
        return vp + apply_along_axis(__ortho, 1, prj)

    @staticmethod
    def __project(vp, geom, prjOp, npts):
        assert isinstance(vp, Point), 'vp is expected to be a Shapely Point!'
        assert GeomLib.isAShapelyGeometry(geom), 'geom is expected to be a Shapely geometry!'
        assert not GeomLib.isMultipart(geom), 'geom must be a single Shapely Point, LineString or Polygon!'

        vpCoord = vp.coords
        if isinstance(geom, Point):
            return Point(prjOp(vpCoord, geom.coords))

        _densifier = PointsDensifierLib3D.densifyByCurvilinearAbscissa
        _curvAbsc = linspace(0, 1, npts)

        if isinstance(geom, LinearRing):
            _tmp = _densifier(geom, curvAbsc=_curvAbsc, blockid=0)
            geom = LinearRing([x['geometry'] for x in _tmp])
            return LinearRing(prjOp(vpCoord, geom.coords))

        if isinstance(geom, LineString):
            _tmp = _densifier(geom, curvAbsc=_curvAbsc, blockid=0)
            geom = LineString([x['geometry'] for x in _tmp])
            return LineString(prjOp(vpCoord, geom.coords))

        if isinstance(geom, Polygon):
            _tmp1 = _densifier(geom.exterior, curvAbsc=_curvAbsc, blockid=0)
            _tmp1 = LinearRing([x['geometry'] for x in _tmp1])

            _tmp2 = [_densifier(hole, curvAbsc=_curvAbsc, blockid=0) for hole in geom.interiors]
            _tmp2 = [LinearRing([x['geometry'] for x in _tmp1]) for _tmp1 in _tmp2]
            geom = Polygon(_tmp1, _tmp2)

            prjExterior = prjOp(vpCoord, geom.exterior.coords)
            prjHoles = [prjOp(vpCoord, hole.coords) for hole in geom.interiors]
            return Polygon(prjExterior, prjHoles)

        raise Exception('Unreachable instruction!')

    @staticmethod
    def orthogonal(vp, geom, npts=51):
        prjOp = SphericalProjectionLib.__orthogonalProjectionOnTheUnitDisc
        return SphericalProjectionLib.__project(vp, geom, prjOp, npts)

    @staticmethod
    def stereographic(vp, geom, npts=51):
        prjOp = SphericalProjectionLib.__stereographicProjectionOnTheUnitDisc
        return SphericalProjectionLib.__project(vp, geom, prjOp, npts)



'''
Created on 5 sept. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import asarray, ceil, mean, pi, sum
from shapely.geometry import LineString, Polygon
from t4gpd.commons.GeomLib3D import GeomLib3D
from t4gpd.pyvista.ToUnstructuredGrid import ToUnstructuredGrid
from t4gpd.pyvista.commons.RayCasting3DLib import RayCasting3DLib
from t4gpd.pyvista.commons.RayCastingIn3DLib import RayCastingIn3DLib

from t4gpd.pyvista.commons.Diameter3DLib import Diameter3DLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class SVF3DLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def isFiveTimeRuleTrue(f1, f2):
        '''
        Wang, F. (2020). Modelling of urban micro-climates for building applications: a 
        non-transparent radiative transfer approach [Univ. de Lyon]. 
        https://tel.archives-ouvertes.fr/tel-03358937
        
        Excerpt from Section 2.2.2.4:
        a finite-area Lambertian emitter should be modeled as a point source only when 
        the distance to the receiving surface is greater than five times the maximum 
        projected width of the emitter
        '''
        assert isinstance(f1, Polygon), 'f1 is expected to be a Shapely Polygon!'
        assert isinstance(f2, Polygon), 'f2 is expected to be a Shapely Polygon!'

        _, diam1 = Diameter3DLib.diameter3D(f1)
        dist12 = GeomLib3D.distFromTo(GeomLib3D.centroid(f1).coords[0], GeomLib3D.centroid(f2).coords[0])
        return (dist12 >= 5 * diam1)

    @staticmethod
    def isa(f1, f2, optim=None):
        if optim is None:
            return SVF3DLib.__infinitesimalSurfaceApproximation(f1, f2)

        if ('left' == optim.lower()):
            F1 = SVF3DLib.subdivideLeft(f1, f2)
            # print(f'\t{len(F1)} sub-faces (left)')
            vfs = [
                SVF3DLib.__infinitesimalSurfaceApproximation(_f1, f2)
                for _f1 in F1
                ]
            return mean(vfs)

        if ('right' == optim.lower()):
            F2 = SVF3DLib.subdivideLeft(f2, f1)
            # print(f'\t{len(F2)} sub-faces (right)')
            vfs = [
                SVF3DLib.__infinitesimalSurfaceApproximation(f1, _f2)
                for _f2 in F2
                ]
            return sum(vfs)

        if ('both' == optim.lower()):
            F1 = SVF3DLib.subdivideLeft(f1, f2)
            F2 = SVF3DLib.subdivideLeft(f2, f1)
            # print(f'\t{len(F1) * len(F2)} sub-faces (both)')
            vfs = [sum([SVF3DLib.__infinitesimalSurfaceApproximation(_f1, _f2)
                for _f2 in F2]) for _f1 in F1]
            return mean(vfs)

        raise IllegalArgumentTypeException(optim, 'None, "left", "right" or "both"')

    @staticmethod
    def __infinitesimalSurfaceApproximation(f1, f2):
        '''
        View Factor from f1 to f2: diffuse energy leaving f1 directly toward and intercepted by f2

        Based on Section 4.2 "Definition of View Factors", (Modest and Mazumder, 2022, p. 150)
        Modest, M. F., and Mazumder, S. (2022). Radiative Heat Transfer. In Academic Press (Ed.), 
        Radiative Heat Transfer (4th Edition). Elsevier. https://doi.org/10.1016/c2018-0-03206-5

        Based on Section 2.2.2.3 of (Wang, 2020)
        Wang, F. (2020). Modelling of urban micro-climates for building applications: a 
        non-transparent radiative transfer approach [Univ. de Lyon]. 
        https://tel.archives-ouvertes.fr/tel-03358937
        '''
        a2 = GeomLib3D.getArea(f2)
        n1, n2 = GeomLib3D.getFaceNormalVector(f1), GeomLib3D.getFaceNormalVector(f2)
        f1Tof2 = GeomLib3D.vector_to(GeomLib3D.centroid(f1).coords[0], GeomLib3D.centroid(f2).coords[0])
        R = GeomLib3D.norm3D(f1Tof2)

        return (-a2 * GeomLib3D.dotProduct(n1, f1Tof2) * GeomLib3D.dotProduct(n2, f1Tof2)) / (pi * R ** 4)

    @staticmethod
    def monteCarlo(f1, f2, nrays, optim=False):
        if not optim:
            return SVF3DLib.__monteCarlo(f1, f2, nrays)

        if optim:
            vfs = [ SVF3DLib.__monteCarlo(f1, f2, nrays) for _ in range(5) ]
            return mean(vfs)
            
        raise IllegalArgumentTypeException(optim, 'True or False')

    @staticmethod
    def __monteCarlo(f1, f2, nrays):
        '''
        View Factor from f1 to f2: diffuse energy leaving f1 directly toward and intercepted by f2
        '''
        masks = [ GeoDataFrame([ {'geometry': f1}, {'geometry': f2} ]) ]
        ugrid = ToUnstructuredGrid(masks).run()
        obbTree = RayCastingIn3DLib.prepareVtkOBBTree(ugrid)
        maxRayLen = 2 * ugrid.length

        n1 = asarray(GeomLib3D.getFaceNormalVector(f1))
        rays = RayCasting3DLib.prepareOrientedRandomRays(nrays, n1, openness=90, method='MonteCarlo')
        shootingDirs = maxRayLen * rays

        epsilon = 1e-6
        srcPt = asarray(GeomLib3D.centroid(f1).coords[0]) + epsilon * n1
        dstPts = srcPt + shootingDirs
        srcPts = srcPt.reshape(1, -1).repeat(len(dstPts), axis=0)

        rays, _, hitDists, _ = RayCastingIn3DLib.mraycastObbTree(obbTree, srcPts, dstPts)

        '''
        # DEBUG
        rows = []
        for i in range(len(rays)):
            if (hitDists[i] != RayCastingIn3DLib.INFINITY):
                row = {
                    'hit': 1 if (hitDists[i] == RayCastingIn3DLib.INFINITY) else 2,
                    'geometry': LineString(rays[i]),
                    }
                rows.append(row)
        gdf1 = GeoDataFrame(data=[ {'gid':1, 'geometry': f1}, {'gid':2, 'geometry': f2} ])
        gdf2 = GeoDataFrame(data=rows)
        scene = ToUnstructuredGrid([gdf1, gdf2], fieldname='gid').run()
        scene.plot(show_edges=True)
        '''

        return len([d for d in hitDists if (d != RayCastingIn3DLib.INFINITY)]) / len(hitDists)

    @staticmethod
    def nusseltUnitSphere(f1, f2):
        '''
        Based on Section 4.9 "The Unit Sphere Method", (Modest and Mazumder, 2022, p. 150)

        Modest, M. F., and Mazumder, S. (2022). Radiative Heat Transfer. In Academic Press (Ed.), 
        Radiative Heat Transfer (4th Edition). Elsevier. https://doi.org/10.1016/c2018-0-03206-5
        '''
        return SVF3DLib.__infinitesimalSurfaceApproximation(f1, f2)

    @staticmethod
    def __subdivide(f1, n):
        coords = list(f1.exterior.coords)
        if (coords[0] == coords[-1]):
            coords = coords[0:-1]
        
        result = []
        if (3 == len(coords)):
            a, b, c = coords
            u = asarray(GeomLib3D.vector_to(a, b)) / n
            v = asarray(GeomLib3D.vector_to(a, c)) / n
            a = asarray(a)
            for l in range(n):
                for c in range(n - l):
                    p = Polygon([
                        a + c * u + l * v,
                        a + (c + 1) * u + l * v,
                        a + c * u + (l + 1) * v,
                        ])
                    result.append(p)
                    if (c + 1) < (n - l):
                        p = Polygon([
                            a + (c + 1) * u + l * v,
                            a + (c + 1) * u + (l + 1) * v,
                            a + c * u + (l + 1) * v,
                            ])
                        result.append(p)
            return result

        elif(4 == len(coords)):
            a, b, c, d = coords
            u = asarray(GeomLib3D.vector_to(a, b)) / n
            v = asarray(GeomLib3D.vector_to(a, d)) / n
            a = asarray(a)
            for l in range(n):
                for c in range(n):
                    p = Polygon([
                        a + c * u + l * v,
                        a + (c + 1) * u + l * v,
                        a + (c + 1) * u + (l + 1) * v,
                        a + c * u + (l + 1) * v,
                        ])
                    result.append(p)
            return result
        raise 'The subdivide(...) method expects a triangle or a rectangle!'

    @staticmethod
    def subdivideLeft(f1, f2):
        if SVF3DLib.isFiveTimeRuleTrue(f1, f2):
            return [f1]
        else:
            _, diam1 = Diameter3DLib.diameter3D(f1)
            dist12 = GeomLib3D.distFromTo(GeomLib3D.centroid(f1).coords[0], GeomLib3D.centroid(f2).coords[0])

            n = int(ceil((5 * diam1) / (dist12)))
            return SVF3DLib.__subdivide(f1, n)



'''
Created on 22 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import log2, roll

from t4gpd.commons.crossroads_identification.AbstractMethod import AbstractMethod


try:
    from pywt import wavedec, waverec

    class FWTMethod(AbstractMethod):
        '''
        classdocs
        '''

        def __init__(self, level=None, wavelet='haar', nRays=64, nOutputRays=8):
            self.level = level
            self.wavelet = wavelet
            self.nRays = nRays
            self.nOutputRays = nOutputRays

        def attrName(self):
            return 'recId_fwt'

        def distance(self, signal, signalRef):
            return sum(abs(signal - signalRef)) / sum(abs(signalRef))

        def nearestPattern(self, shapeSignature, patternSignatures):
            minDist, minCircShift, closestPatternID = float('inf'), None, None

            for gid in list(patternSignatures.keys()):
                for circShift in list(patternSignatures[gid].keys()):  # SYSTEMATIC ROTATION
                    patternSignature = patternSignatures[gid][circShift]
                    _dist = self.distance(shapeSignature, patternSignature)
                    if _dist < minDist:
                        minDist = _dist
                        minCircShift = circShift
                        closestPatternID = gid
            rotation = -(360.0 * minCircShift) / float(self.nRays)
            return {
                self.attrName():closestPatternID,
                'rotation': rotation,
                'min_dist': minDist
                }

        def __signature(self, raylens):
            coeffs = wavedec(raylens, self.wavelet, level=self.level)
            cAn, cDn = coeffs[0:2]
            rec = waverec([cAn, cDn], self.wavelet)
            return rec

        def signature(self, raylens, isAPatternLayer):
            nRays = len(raylens)
            if self.level is None:
                self.level = int(log2(nRays / self.nOutputRays))

            if isAPatternLayer:
                signatures = dict()
                for circShift in range(0, nRays):  # SYSTEMATIC ROTATION
                # for circShift in [0]:
                    _raylens = roll(raylens, circShift)
                    signatures[circShift] = self.__signature(_raylens)
                return signatures
            return self.__signature(raylens)

except ImportError:

    class FWTMethod(AbstractMethod):
        '''
        classdocs
        '''

        def __init__(self, level=None, wavelet='haar', nRays=64, nOutputRays=8):
            raise Exception('Install PyWavelets first!')



'''
Created on 15 sept. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from _warnings import warn
from datetime import datetime
import re

from pandas import read_csv
from t4gpd.commons.GeoProcess import GeoProcess


class SensirionReader(GeoProcess):
    '''
    classdocs
    '''
    FIELD_NAMES = ['local_datetime', 'Tair', 'RH'] 

    RE1 = re.compile(r'^# EdfVersion=(\d\.\d)$')
    RE2 = re.compile(r'^# SensorFamily=(\w*).*$')
    RE3 = re.compile(r'^# SensorId=(.*)$')

    def __init__(self, inputFile):
        '''
        Constructor
        '''
        self.inputFile = inputFile

    def __getEdfVersionAndSensorSpecs(self):
        with open(self.inputFile, 'r') as f:
            for nline, line in enumerate(f, start=1):
                line = line.strip()
                if (1 == nline):
                    version = float(self.RE1.search(line).group(1))
                elif (6 == nline):
                    sensorFamily = self.RE2.search(line).group(1)
                elif (7 == nline):
                    sensorId = self.RE3.search(line).group(1)
                    sensorId = sensorId.replace(':', '')
                    return version, sensorFamily, sensorId

    def run(self):
        # https://sensirion.com/products/catalog/SHT40/
        version, sensorFamily, sensorId = self.__getEdfVersionAndSensorSpecs()

        if (4.0 != version):
            warn('EdfVersion is expected to be equal to 4.0!')

        # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
        df = read_csv(self.inputFile, header=None, names=self.FIELD_NAMES,
                      parse_dates=['local_datetime'], sep='\s+',
                      skiprows=10, usecols=range(1, 4))
        # df.Epoch_UTC = df.Epoch_UTC.apply(lambda v: datetime.fromtimestamp(v))
        df.rename(columns={'local_datetime': 'timestamp'}, inplace=True)
        df['sensorFamily'] = sensorFamily
        df['sensorId'] = sensorId
        df['station'] = f'{sensorFamily}-{sensorId}'

        return df

'''
inputFile = '/home/tleduc/prj/nm-ilots-frais/terrain/220711/Sensirion_MyAmbience_SHT40_Gadget_0F0E_2022-07-11T18-49-50.435746.edf'
df = SensirionReader(inputFile).run()
print(df.head(5))
'''



'''
Created on 10 juin 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import clip
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import box, MultiPolygon
from shapely.ops import unary_union
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.RayCasting3Lib import RayCasting3Lib
from t4gpd.commons.KernelLib import KernelLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class CrossroadsStarDomain(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, buildings, nRays=64, maxRayLength=100, debug=False):
        '''
        Constructor
        '''
        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(buildings, 'GeoDataFrame')
        self.buildings = buildings
        # CLEAN GEOMETRIES
        self.buildings.geometry = self.buildings.geometry.apply(lambda g: g.buffer(0))
        self.buildingsIdx = buildings.sindex

        # self.shootingDirs = RayCastingLib.preparePanopticRays(nRays)
        self.shootingDirs = RayCasting3Lib.preparePanopticRays(nRays)
        self.maxRayLength = maxRayLength
        self.debug = debug

    def __assessKernel(self, geom, r):
        if (self.minRayLength >= r):
            _kernel = _open_spaces = geom.buffer(self.minRayLength)
            if self.debug:
                self.debugList.append(_open_spaces)
                self.debugRayLens.append(round(self.minRayLength, 1))
            return _kernel
        
        _open_spaces = geom.buffer(r)

        _buildings = clip(self.buildings, box(*_open_spaces.bounds))
        if (0 == len(_buildings)):
            return _open_spaces
        _buildings = GeoDataFrame([{'geometry': unary_union(_buildings.geometry)}],
                                  crs=self.buildings.crs)
        _buildings = _buildings.explode(ignore_index=True)
        _buildings = _buildings[ ~_buildings.geometry.isna() ]
        _buildings.geometry = _buildings.geometry.apply(lambda g: g.convex_hull)
        _buildings = unary_union(_buildings.geometry)

        _open_spaces = _open_spaces.difference(_buildings)
        _kernel, flag = KernelLib.getKernel(_open_spaces)

        if self.debug:
            self.debugList.append(_open_spaces)
            self.debugRayLens.append(round(r, 1))

        if (flag and (geom.within(_kernel))):
            return _kernel

        return None

    def __assessKernelByDichotomy(self, geom, r1, r2):
        if (r2 - r1 < 1.0):
            if (0 < r1):
                return self.__assessKernel(geom, r1)
            return None

        r12 = (r1 + r2) / 2.0
        _kernel = self.__assessKernel(geom, r12)
        if (_kernel is None):
            return self.__assessKernelByDichotomy(geom, r1, r12)
        else:
            return self.__assessKernelByDichotomy(geom, r12, r2)

    def runWithArgs(self, row):
        geom = row.geometry.centroid

        if self.debug:
            self.debugList = []
            self.debugRayLens = []

        buffDist = 40.0
        self.minRayLength, _, _ = GeomLib.getNearestFeature(
            self.buildings, self.buildingsIdx, geom, buffDist)

        # _, _, hitDists, _ = RayCastingLib.multipleRayCast2D(
        #     self.buildings, self.buildingsIdx, geom, self.shootingDirs, self.maxRayLength)
        _, _, hitDists = RayCasting3Lib.outdoorMultipleRayCast2D(
            self.buildings, geom, self.shootingDirs, self.maxRayLength)
        self.maxRayLength = max(hitDists)

        kernel = None
        if (0 < self.minRayLength):
            kernel = self.__assessKernelByDichotomy(geom, 0, self.maxRayLength)
        if kernel is None:
            kernel = geom

        if self.debug:
            return {
                'geometry': MultiPolygon(self.debugList),
                'RayLens': ArrayCoding.encode(self.debugRayLens)
                }

        return {
            'geometry': kernel,
            'MinLenRad': self.minRayLength,
            'MaxLenRad': self.maxRayLength,
            'kern_drift': geom.distance(kernel.centroid)
            }



'''
Created on 15 sept. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.commons.GeoProcess import GeoProcess
from pandas import read_csv
from _warnings import warn


class CampbellSciReader(GeoProcess):
    '''
    classdocs
    '''
    FIELD_NAMES = [
        'timestamp', 'RECORD', 'BattV_Min', 'PTemp_C_Avg', 'CDM1BattV_Min', 'CDM1PTempC1',
        'CDM1PTempC2', 'CDM1PTempC3', 'CDM1PTempC4', 'AirTC_Avg', 'AirTC_Min', 'AirTC_Max',
        'RH_Avg', 'RH_Min', 'RH_Max', 'WindDir', 'WS_ms_Avg', 'WS_ms', 'WS_ms_Min',
        'WS_ms_Max', 'WS_ms_S_WVT', 'WindDir_D1_WVT', 'WSDiag', 'SmplsF_Tot', 'Diag1F_Tot',
        'Diag2F_Tot', 'Diag4F_Tot', 'Diag8F_Tot', 'Diag9F_Tot', 'Diag10F_Tot', 'NNDF_Tot',
        'CSEF_Tot', 'SR01Up_1_Avg', 'SR01Dn_1_Avg', 'IR01Up_1_Avg', 'IR01Dn_1_Avg',
        'NR01TC_1_Avg', 'NR01TK_1_Avg', 'NetRs_1_Avg', 'NetRl_1_Avg', 'Albedo_1_Avg',
        'UpTot_1_Avg', 'DnTot_1_Avg', 'NetTot_1_Avg', 'IR01UpCo_1_Avg', 'IR01DnCo_1_Avg',
        'SR01Up_2_Avg', 'SR01Dn_2_Avg', 'IR01Up_2_Avg', 'IR01Dn_2_Avg', 'NR01TC_2_Avg',
        'NR01TK_2_Avg', 'NetRs_2_Avg', 'NetRl_2_Avg', 'Albedo_2_Avg', 'UpTot_2_Avg',
        'DnTot_2_Avg', 'NetTot_2_Avg', 'IR01UpCo_2_Avg', 'IR01DnCo_2_Avg', 'SR01Up_3_Avg',
        'SR01Dn_3_Avg', 'IR01Up_3_Avg', 'IR01Dn_3_Avg', 'NR01TC_3_Avg', 'NR01TK_3_Avg',
        'NetRs_3_Avg', 'NetRl_3_Avg', 'Albedo_3_Avg', 'UpTot_3_Avg', 'DnTot_3_Avg',
        'NetTot_3_Avg', 'IR01UpCo_3_Avg', 'IR01DnCo_3_Avg', 'Temp_C_Avg(1)',
        'Temp_C_Max(1)', 'Temp_C_Min(1)', 'Temp_C(1)', 'Temp_C_Std(1)', 'Temp_C_Avg(2)',
        'Temp_C_Max(2)', 'Temp_C_Min(2)', 'Temp_C(2)', 'Temp_C_Std(2)']

    FIELDS_TO_BE_DROPPED = [
        "RECORD", "BattV_Min", "PTemp_C_Avg", "CDM1BattV_Min", "CDM1PTempC1",
        "CDM1PTempC2", "CDM1PTempC3", "CDM1PTempC4", "AirTC_Min", "AirTC_Max", "RH_Min",
        "RH_Max", "WS_ms", "WS_ms_Min", "WS_ms_Max", "WS_ms_S_WVT", "WindDir_D1_WVT",
        "WSDiag", "SmplsF_Tot", "Diag1F_Tot", "Diag2F_Tot", "Diag4F_Tot", "Diag8F_Tot",
        "Diag9F_Tot", "Diag10F_Tot", "NNDF_Tot", "CSEF_Tot", "IR01Up_1_Avg", "IR01Dn_1_Avg",
        "NR01TC_1_Avg", "NR01TK_1_Avg", "NetRs_1_Avg", "NetRl_1_Avg", "UpTot_1_Avg",
        "DnTot_1_Avg", "NetTot_1_Avg", "IR01Up_2_Avg", "IR01Dn_2_Avg", "NR01TC_2_Avg",
        "NR01TK_2_Avg", "NetRs_2_Avg", "NetRl_2_Avg", "UpTot_2_Avg", "DnTot_2_Avg",
        "NetTot_2_Avg", "IR01Up_3_Avg", "IR01Dn_3_Avg", "NR01TC_3_Avg", "NR01TK_3_Avg",
        "NetRs_3_Avg", "NetRl_3_Avg", "UpTot_3_Avg", "DnTot_3_Avg", "NetTot_3_Avg",
        "Temp_C_Max(1)", "Temp_C_Min(1)", "Temp_C(1)", "Temp_C_Std(1)", "Temp_C_Max(2)",
        "Temp_C_Min(2)", "Temp_C(2)", "Temp_C_Std(2)"]

    def __init__(self, inputFile):
        '''
        Constructor
        '''
        self.inputFile = inputFile

    def run(self):
        # https://www.campbellsci.com/cr1000x
        df = read_csv(self.inputFile, header=None, names=self.FIELD_NAMES,
                      na_values='NAN', parse_dates=[0], sep=',', skiprows=4)

        dfUnits = read_csv(self.inputFile, header=None, names=self.FIELD_NAMES,
                           nrows=1, sep=',', skiprows=2)

        for fieldname in ['WS_ms_Avg', 'WS_ms', 'WS_ms_Min', 'WS_ms_Max', 'WS_ms_S_WVT']:
            if ('meters/second' != dfUnits[fieldname].squeeze().lower()):
                warn(f'"{fieldname}" is expected to be in m/s!')

        df['sensorFamily'] = 'cr1000x'
        df.drop(columns=self.FIELDS_TO_BE_DROPPED, inplace=True)

        return df

'''
inputFile = '/home/tleduc/prj/nm-ilots-frais/terrain/220711/CR1000XSeries_TwoSec.dat'
df = CampbellSciReader(inputFile).run()
print(df.head(5))
'''



'''
Created on 22 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry.multilinestring import MultiLineString
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException

from t4gpd.commons.crossroads_identification.FFTMethod import FFTMethod
from t4gpd.commons.crossroads_identification.FWTMethod import FWTMethod
from t4gpd.commons.crossroads_identification.MeanAngularityMethod import MeanAngularityMethod
from t4gpd.commons.crossroads_identification.MeanVectorMethod import MeanVectorMethod


class CrossroadRecognitionLib(object):
    '''
    classdocs
    '''

    # def __init__(self, identificationMethod, nRays, maxRayLength=None):
    def __init__(self, **kwargs):
        '''
        Constructor
        '''
        method = kwargs['method'] if ('method' in kwargs) else None 

        if method in ['meth1', 'MeanVector']:
            self.method = MeanVectorMethod()

        elif method in ['meth2', 'FFT', 'Fourier']:
            self.method = FFTMethod()

        elif method in ['meth3', 'MeanAngularity']:
            if ('maxRayLength' in kwargs):
                maxRayLength = kwargs['maxRayLength']
            else:
                raise Exception('Set the value of "maxRayLength"!')
            maxThresholdRatio = kwargs['maxThresholdRatio'] if ('maxThresholdRatio' in kwargs) else 0.9
            self.method = MeanAngularityMethod(maxRayLength, maxThresholdRatio)

        elif method in ['meth4', 'FWT', 'Wavelet']:
            if ('nRays' in kwargs):
                nRays = kwargs['nRays']
            else:
                raise Exception('Set the value of "nRays"!')
            wavelet = kwargs['wavelet'] if ('wavelet' in kwargs) else 'Haar'
            nOutputRays = kwargs['nOutputRays'] if ('nOutputRays' in kwargs) else 8
            self.method = FWTMethod(None, wavelet, nRays, nOutputRays)

        else:
            raise Exception('Illegal argument: %s is not a valid method!' % method)

    def attrName(self):
        return self.method.attrName()

    def nearestPattern(self, shapeSignature, patternSignatures):
        return self.method.nearestPattern(shapeSignature, patternSignatures)

    def signatures(self, inputGdf, idFieldName, isAPatternLayer):
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        if idFieldName not in inputGdf:
            raise Exception('%s is not a relevant field name!' % (idFieldName))

        signatures = dict()
        for _, row in inputGdf.iterrows():
            if not isinstance(row.geometry, MultiLineString):
                raise Exception('Must be a GeoDataFrame of MultiLineString!')
            geom = row.geometry
            gid = row[idFieldName]
            raylens = GeomLib.fromMultiLineStringToLengths(geom)
            signatures[gid] = self.method.signature(raylens, isAPatternLayer) 
        return signatures

    def signature(self, raylens, isAPatternLayer=False):
        return self.method.signature(raylens, isAPatternLayer)
        



'''
Created on 26 oct. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import timedelta


class DataFrameLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def interpolate(df, xfieldname, yfieldname, x):
        assert (df[xfieldname].is_monotonic_increasing or 
                df[xfieldname].is_monotonic_decreasing), \
                f'Column "{xfieldname}" must be monotonic'

        df.set_index(xfieldname, drop=False, inplace=True)

        row0 = df.iloc[ df.index.get_indexer([x], method='ffill') ]
        row1 = df.iloc[ df.index.get_indexer([x], method='bfill') ]

        x0, x1 = row0[xfieldname].squeeze(), row1[xfieldname].squeeze()
        y0, y1 = row0[yfieldname].squeeze(), row1[yfieldname].squeeze()
        delta, deltaX = x - x0, x1 - x0

        if isinstance(deltaX, timedelta):
            delta = delta.seconds
            deltaX = deltaX.seconds

        if (0 == deltaX):
            return y0
        return y0 + (delta * (y1 - y0)) / deltaX



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import arctan2, cos, pi, sin
from shapely.geometry import Point, Polygon
from shapely.ops import unary_union


class SequenceRadii(object):
    '''
    classdocs
    '''

    def __init__(self, nbranchs=8, width=12.0, length=100.0):
        '''
        Constructor
        '''
        self.nbranchs = nbranchs
        self.width = 0.5 * float(width)
        self.length = float(length)
        # self.alpha = arcsin(self.width / self.length)
        self.alpha = 2 * arctan2(self.width, self.length)
        self.angles = [((2.0 * pi * i) / nbranchs) for i in range(nbranchs)]
        self.mainDirs = [(cos(angle), sin(angle)) for angle in self.angles]
        self.incAngle = self.alpha / 2.0

    def getBranch(self, i, centre=[0, 0]):
        w, hw, l = self.width, 0.0, (self.length * cos(self.alpha))
        cx, cy = centre
        ux, uy = self.mainDirs[i]
        vx, vy = -uy, ux
        p0 = (cx + hw * ux - w * vx, cy + hw * uy - w * vy)
        p1 = (cx + l * ux - w * vx, cy + l * uy - w * vy)
        p2 = (cx + l * ux + w * vx, cy + l * uy + w * vy)
        p3 = (cx + hw * ux + w * vx, cy + hw * uy + w * vy)
        return Polygon([p0, p1, p2, p3, p0])

    def getSector(self, sector, centre=[0, 0]):
        i, j = sector

        if ((0 == i) and (self.nbranchs == j)):
            result = Point(centre).buffer(self.length, -1)
        else:
            angle0, angle1 = self.angles[i], self.angles[j]
            angle1 = angle1 + 2.0 * pi if (angle0 > angle1) else angle1
            l = self.length * cos(self.alpha)
            cx, cy = centre

            result = [centre]
            angle = angle0
            while angle < angle1:
                p = (cx + cos(angle) * l, cy + sin(angle) * l)
                result.append(p)
                angle += self.incAngle
            result.append(centre)

            br1 = self.getBranch(i, centre)
            br2 = self.getBranch(j, centre)
            polygons = [br1, br2, Polygon(result)]

            # Use a buffer to avoid slivers
            result = unary_union(polygons).buffer(0.001, -1)

        return result



'''
Created on 25 juin 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.AngleLib import AngleLib
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.graph.UrbanGraphLibOld import UrbanGraphLibOld
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class CrossroadsAngularity(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, roads, sampleDist=None, threshold=10):
        '''
        Constructor
        '''
        if not isinstance(roads, GeoDataFrame):
            raise IllegalArgumentTypeException(roads, 'GeoDataFrame')
        self.roads = roads

        self.graph = UrbanGraphLibOld()
        self.graph.add([_geom for _geom in roads.geometry])
        self.rs = self.graph.getUniqueRoadsSections()
        self.assocTbl = dict()
        for rs in self.graph.getUniqueRoadsSections():
            _start = UrbanGraphLibOld.hashCoord(rs['geometry'].coords[0])
            if (_start in self.assocTbl):
                self.assocTbl[_start].append(rs['geometry'].coords)
            else:
                self.assocTbl[_start] = [rs['geometry'].coords]

            _stop = UrbanGraphLibOld.hashCoord(rs['geometry'].coords[-1])
            if (_stop in self.assocTbl):
                self.assocTbl[_stop].append(list(reversed(rs['geometry'].coords)))
            else:
                self.assocTbl[_stop] = [list(reversed(rs['geometry'].coords))]

        self.sampleDist = sampleDist
        self.threshold = threshold

    def __isAMultipleOf(self, a, b):
        if (self.threshold >= (a % b)) or (self.threshold >= (b - (a % b))):
            return True
        return False

    def __aux1(self, currCoords):
        _currHash = UrbanGraphLibOld.hashCoord(currCoords)

        _azims = []
        for _branch in self.assocTbl[_currHash]:
            _remoteCoords = _branch[-1]
            _currDir = [_remoteCoords[0] - currCoords[0], _remoteCoords[1] - currCoords[1]]
            _currAzim = round(AngleLib.toDegrees(AngleLib.normAzimuth(_currDir)), 1)
            _azims.append(_currAzim)

        _azims = sorted(_azims)
        _angles = [round((_azims[i] - _azims[i - 1] + 360) % 360, 1) for i in range(len(_azims))]
        _mode = 4
        for _angle in _angles:
            if (not self.__isAMultipleOf(_angle, 90)):
                if (not self.__isAMultipleOf(_angle, 45)):
                    _mode = 16
                elif (4 == _mode):
                    _mode = 8

        return {
            'cross_azim': ArrayCoding.encode(_azims),
            'cross_std1': None,
            'cross_angl': ArrayCoding.encode(_angles),
            'cross_std2': None,
            'cross_mode': _mode
            }

    def runWithArgs(self, row):
        geom = row.geometry

        if ('Point' == geom.geom_type):
            _currCoords = geom.coords[0]

            if ((self.sampleDist is None) or (0 >= self.sampleDist)):
                return self.__aux1(_currCoords)
            else:
                raise NotImplementedError('Must be implemented!')

        return {
            'cross_azim': None,
            'cross_std1': None,
            'cross_angl': None,
            'cross_std2': None,
            'cross_mode': None
            }



'''
Created on 15 sept. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from glob import glob
from os.path import basename, exists
from sys import stderr

from geopandas import GeoDataFrame
from pandas import concat
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.picoclim.CampbellSciReader import CampbellSciReader
from t4gpd.picoclim.InertialMeasureReader import InertialMeasureReader
from t4gpd.picoclim.KestrelReader import KestrelReader
from t4gpd.picoclim.MeteoFranceReader import MeteoFranceReader
from t4gpd.picoclim.SensirionReader import SensirionReader
from t4gpd.picoclim.TracksWaypointsReader import TracksWaypointsReader


class MetrologicalCampaignReader(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, dirName, dirNameMeteoFr=None):
        '''
        Constructor
        '''
        if not exists(dirName):
            raise IllegalArgumentTypeException(dirName, 'valid directory name')
        self.dirName = dirName

        if not ((dirNameMeteoFr is None) or exists(dirNameMeteoFr)):
            raise IllegalArgumentTypeException(dirNameMeteoFr, 'valid directory name')
        self.dirNameMeteoFr = dirNameMeteoFr

    def __loader(self, filenameFilter, Reader):
        dfs = []
        for filename in glob(filenameFilter):
            print(f'*** {basename(filename)}', file=stderr)
            df = Reader(filename).run()
            dfs.append(df)
        return None if (0 == len(dfs)) else concat(dfs, ignore_index=True)

    def run(self):
        filename = f'{self.dirName}/sig.gpkg'
        print(f'*** {basename(filename)}', file=stderr)
        static, tracks, waypoints = TracksWaypointsReader(filename).run()

        filenameFilter = f'{self.dirName}/LOG_FILE_????????_*.txt'
        dfImu = self.__loader(filenameFilter, InertialMeasureReader)
        dfImu.sort_values(by='timestamp', inplace=True, ignore_index=True)

        filenameFilter = f'{self.dirName}/CR1000XSeries_TwoSec.dat'
        dfMob = self.__loader(filenameFilter, CampbellSciReader)
        if dfMob is not None:
            dfMob.sort_values(by='timestamp', inplace=True, ignore_index=True)

        filenameFilter = f'{self.dirName}/*HEAT*.csv'
        dfStat1 = self.__loader(filenameFilter, KestrelReader)
        if dfStat1 is not None:
            dfStat1.sort_values(by='timestamp', inplace=True, ignore_index=True)

        filenameFilter = f'{self.dirName}/Sensirion_*.edf'
        dfStat2 = self.__loader(filenameFilter, SensirionReader)
        if dfStat2 is not None:
            dfStat2.sort_values(by='timestamp', inplace=True, ignore_index=True)
            if static is not None:
                dfStat2 = dfStat2.merge(static, on='station')
                dfStat2 = GeoDataFrame(dfStat2, crs=static.crs)

        filenameFilter = f'{self.dirNameMeteoFr}/meteo-france-*.txt'
        dfMeteoFr = self.__loader(filenameFilter, MeteoFranceReader)
        if dfMeteoFr is not None:
            dfMeteoFr.sort_values(by='timestamp', inplace=True, ignore_index=True)

        return static, tracks, waypoints, dfImu, dfMob, dfStat1, dfStat2, dfMeteoFr

'''
dirName1 = '/home/tleduc/prj/nm-ilots-frais/terrain/220711'
# dirName1 = '/home/tleduc/prj/nm-ilots-frais/terrain/220713'
dirName2 = '/home/tleduc/prj/nm-ilots-frais/information'

static, tracks, waypoints, dfImu, dfMob, dfStat1, dfStat2, dfMeteoFr = \
    MetrologicalCampaignReader(dirName1, dirName2).run()
'''



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from itertools import combinations

from numpy import ceil, log10

from t4gpd.commons.crossroads_generation.Sequence import Sequence


class SequencesGeneration(object):
    '''
    classdocs
    '''
    def __init__(self, nbranchs=8, mirror=False, withBranchs=True, withSectors=True):
        '''
        Constructor
        '''
        self.nbranchs = nbranchs
        self.mirror = mirror
        self.withBranchs = withBranchs
        self.withSectors = withSectors

    def __addSingleSeq(self, intermed, seq):
        s = Sequence(self.nbranchs, seq, self.mirror)
        if not s in intermed.keys():
            intermed[s] = None

    def __addMultipleSeq(self, intermed, sector, listOfBranchs):
        for n in range(0, 1 + len(listOfBranchs)):
            listOfCombinations = list(combinations(listOfBranchs, n))
            for combins in listOfCombinations:
                seq = [sector]
                for item in combins:
                    seq.append(item)
                self.__addSingleSeq(intermed, seq)

    def __getSequencesOfBranches(self):
        intermed = dict()
        for n in range(1, self.nbranchs + 1):
            intermed[n] = dict()
            for seq in combinations(range(self.nbranchs), n):
                self.__addSingleSeq(intermed[n], seq)

        maxNumOfSeqPerLen = max([len(intermed[n]) for n in intermed.keys()])
        magnitude = 10 ** int(ceil(log10(maxNumOfSeqPerLen)))

        result = dict()
        for n in intermed.keys():
            for i, seqRef in enumerate(intermed[n].keys()):
                result[n * magnitude + i] = seqRef

        return result

    def __getSequencesOfSectors(self):
        intermed = dict()

        if (0 == (self.nbranchs % 4)):
            q1 = int(self.nbranchs / 4)

            for i in [1, 2, 3]:
                q = i * q1
                complement = [j for j in range(self.nbranchs) if (q < j)]
                intermed[i] = dict()
                self.__addMultipleSeq(intermed[i], (0, q), complement)

            maxNumOfSeqPerLen = max([len(intermed[n]) for n in intermed.keys()])
            magnitude = 10 ** int(ceil(log10(maxNumOfSeqPerLen)))          

        result = dict()
        for n in intermed.keys():
            for i, seqRef in enumerate(intermed[n].keys()):
                result[-n * magnitude - i] = seqRef

        result[-4 * magnitude] = Sequence(self.nbranchs, [(0, self.nbranchs)], self.mirror)

        return result

    def run(self):
        result = dict()

        if (self.withBranchs and self.withSectors):
            allSeq = [self.__getSequencesOfBranches(), self.__getSequencesOfSectors()]
        elif self.withBranchs:
            allSeq = [self.__getSequencesOfBranches()]
        elif self.withSectors:
            allSeq = [self.__getSequencesOfSectors()]
        else:
            allSeq = []

        for r in allSeq:
            result.update(r)
        return result



'''
Created on 15 sept. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from os.path import basename

from dateutil.parser import parse
from geopandas import GeoDataFrame
from numpy.random import choice
from pandas import read_csv
from shapely.geometry import Point
from t4gpd.commons.GeoProcess import GeoProcess


class InertialMeasureReader(GeoProcess):
    '''
    classdocs
    '''
    FIELD_NAMES = [
        'timestamp', 'step_count', 'X', 'Y', 'Z', 'Distance', 'degree',
        'latitude', 'longitude', 'GpsAccuracy', 'indoor_outdoor_flag', 'TagName'
    ]

    def __init__(self, inputFile, outputCrs=None):
        '''
        Constructor
        '''
        self.inputFile = inputFile
        self.outputCrs = outputCrs

    def __extract_YYYYMMDD_from_filename(self):
        yyyymmdd = basename(self.inputFile)[9:17]
        year, month, day = yyyymmdd[0:4], yyyymmdd[4:6], yyyymmdd[6:8]
        return year, month, day

    def __getRndSubtrackId(self):
        return ''.join(choice([str(i) for i in range(10)], size=10))

    def run(self):
        df = read_csv(self.inputFile, decimal=',', header=None,
                      names=self.FIELD_NAMES, sep='\s+',
                      skip_blank_lines=True, skiprows=2)

        year, month, day = self.__extract_YYYYMMDD_from_filename()
        df.timestamp = df.timestamp.apply(lambda t: parse(f'{year}-{month}-{day}T{t}'))

        df['subtrack'] = self.__getRndSubtrackId()

        df['geometry'] = list(zip(df.longitude, df.latitude))
        df.geometry = df.geometry.apply(lambda t: Point(t))
        df = GeoDataFrame(df, crs='epsg:4326')

        if self.outputCrs is None:
            return df
        return df.to_crs(self.outputCrs)

'''
inputFile = '/home/tleduc/prj/nm-ilots-frais/terrain/220711/LOG_FILE_20220711_172000.txt'
df = InertialMeasureReader(inputFile, outputCrs='epsg:2154').run()
print(df.head(5))
'''



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from shapely.geometry import Point
from shapely.ops import unary_union


class Sequence(object):
    '''
    classdocs
    '''

    def __init__(self, nbranchs=8, seq=None, mirror=False):
        '''
        Constructor
        '''
        self.nbranchs = nbranchs
        self.seq = sorted(seq,
                          key=lambda item: item[0] if self.__isContinuousItem(item) else item,
                          reverse=False)
        self.__mirror = mirror
        self.__validation()

    def __eq__(self, obj):
        if ((obj is None) or 
            (not isinstance(obj, Sequence))):
            return False
        return self.__isCongruent(obj)

    def __hash__(self):
        return len(self.seq)

    def mirror(self):
        mirrorSeq = list()
        for item in self.seq:
            if self.__isContinuousItem(item):
                mirrorSeq.append(((self.nbranchs - item[1]) % self.nbranchs,
                                  (self.nbranchs - item[0]) % self.nbranchs))
            else:
                mirrorSeq.append((self.nbranchs - item) % self.nbranchs)
        return Sequence(self.nbranchs, mirrorSeq, self.__mirror);

    def __isCongruent(self, otherSequence):
        if ((otherSequence is None) or
            (not isinstance(otherSequence, Sequence))):
            return False
        if self.__mirror:
            return (self.__isCongruentModuloARotation(otherSequence) or
                    self.__isCongruentModuloARotation(otherSequence.mirror()))
        else:
            return self.__isCongruentModuloARotation(otherSequence)

    def __isCongruentModuloARotation(self, otherSequence):
        if (self.__strictEquality(otherSequence)):
            return True
        for i in range(1, self.nbranchs):
            if (self.__strictEquality(otherSequence.rotate(i))):
                return True
        return False

    def __isContinuousItem(self, item):
        return (isinstance(item, list) or isinstance(item, tuple))

    def __isValidItem(self, item):
        if self.__isContinuousItem(item):
            return (((0 == item[0]) and (self.nbranchs == item[1])) or 
                    ((item[0] in range(self.nbranchs)) and (item[1] in range(self.nbranchs)))) 
        return (item in range(self.nbranchs))

    def __plus(self, item, offset=1):
        if self.__isContinuousItem(item):
            return ((item[0] + offset) % self.nbranchs, (item[1] + offset) % self.nbranchs)
        return (item + offset) % self.nbranchs

    def __repr__(self):
        return str(self.seq)  # return str(self.seq)

    def __strictEquality(self, otherSequence):
        if ((otherSequence is None) or
            (not isinstance(otherSequence, Sequence)) or
            (self.nbranchs != otherSequence.nbranchs) or
            (len(self.seq) != len(otherSequence.seq))):
            return False
        for i in self.seq:
            if not i in otherSequence.seq:
                return False
        return True

    def __validation(self):
        if (self.nbranchs < 4):
            raise Exception('Illegal argument: nbranchs mut be greater or equal 4!')
        if (len(self.seq) > self.nbranchs):
            raise Exception('Illegal argument: nbranchs mut be greater or equal the seq list number!')
        for item in self.seq:
            if not self.__isValidItem(item):
                raise Exception('Illegal argument: %s is neither in (0..%d) nor in (0..%d, 0..%d)!' % (
                        item, self.nbranchs - 1, self.nbranchs - 1, self.nbranchs - 1))
        # A CONTINUER

    def rotate(self, offset=1):
        newseq = [self.__plus(item, offset) for item in self.seq]
        return Sequence(self.nbranchs, newseq, self.__mirror)

    def asPolygon(self, sequenceRadii, width, centre=[0, 0]):
        polygons = list()
        polygon = Point(centre).buffer(0.5 * width, -1)
        polygons.append(polygon)

        for item in self.seq:
            if self.__isContinuousItem(item):
                polygon = sequenceRadii.getSector(item, centre)
            else:
                polygon = sequenceRadii.getBranch(item, centre)
            polygons.append(polygon)
        result = unary_union(polygons)
        return result

    def getMinModel(self):
        if (4 == self.nbranchs):
            return 4

        if (0 == (self.nbranchs % 4)):
            q1 = int(self.nbranchs / 4)
            setToTest = [0, q1, 2 * q1, 3 * q1]
            if all([(self.__isContinuousItem(i) or (i in setToTest)) for i in self.seq]):
                return 4

            if (8 == self.nbranchs):
                return 8

            if (0 == (self.nbranchs % 8)):
                e1 = int(self.nbranchs / 8)
                setToTest = [0, e1, 2 * e1, 3 * e1, 4 * e1, 5 * e1, 6 * e1, 7 * e1]
                if all([(self.__isContinuousItem(i) or (i in setToTest)) for i in self.seq]):
                    return 8

        return self.nbranchs
